(* $Id: gradient.ml,v 1.12 2014/02/27 16:36:36 fclement Exp $ *)

open Ugraphics;;

type rectangle_gradient_mode =
   | Rect_Horizontal of Ugraphics.from_color * Ugraphics.to_color
   | Rect_Vertical of from_color * to_color
   | Rect_Diagonal1 of from_color * to_color
   | Rect_Diagonal2 of from_color * to_color
   | Rect_Radial_Rectangular of
       from_color * to_color *
       Ugraphics.x * Ugraphics.y * Ugraphics.aspect_ratio
   | Rect_Radial_Elliptic of from_color * to_color * x * y * aspect_ratio
;;

type ellipse_gradient_mode =
   | Ellipse_Horizontal of from_color * to_color
   | Ellipse_Vertical of from_color * to_color
   | Ellipse_Diagonal1 of from_color * to_color
   | Ellipse_Diagonal2 of from_color * to_color
   | Ellipse_Radial_Rectangular of from_color * to_color * x * y * aspect_ratio
   | Ellipse_Radial_Elliptic of from_color * to_color * x * y * aspect_ratio
;;

type arc_gradient_mode =
   | Arc_Horizontal of from_color * to_color
   | Arc_Vertical of from_color * to_color
   | Arc_Diagonal1 of from_color * to_color
   | Arc_Diagonal2 of from_color * to_color
   | Arc_Circular of from_color * to_color
;;

open Unatural;;

(* Horizontal gradient inside a rectangle *)
let grad_rect_h c1 c2 x y w h =
  let rec loop i =
    if i < w then begin
      set_color (scale_color i w c1 c2);
      fill_rect (x + i) y _1_ h;
      loop (succ i);
    end in
  loop zero
;;

(* Vertical gradient inside a rectangle *)
let grad_rect_v c1 c2 x y w h =
  let rec loop i =
    if i < h then begin
      set_color (scale_color i h c1 c2);
      fill_rect x (y + i) w _1_;
      loop (succ i);
    end in
  loop zero
;;

(* First bissector diagonal gradient inside a rectangle *)
let grad_rect_d1 c1 c2 x y w h =
  let sc = w + h in
  let limx = x + w - _1_ in
  let limy = y + h - _1_ in
  let rec loop i x0 y0 x1 y1 =
    set_color (scale_color i sc c1 c2);
    moveto x0 y0;
    if i <= sc then lineto x1 y1;
    if x0 < limx then
      if y1 < limy then
        loop (succ i) (succ x0) y0 x1 (succ y1) else
        loop (succ i) (succ x0) y0 (succ x1) y1 else
    if y0 < limy then
      if y1 < limy then
        loop (succ i) x0 (succ y0) x1 (succ y1) else
        loop (succ i) x0 (succ y0) (succ x1) y1 else
    () in
  set_line_width _1_;
  loop zero x y x y
;;

(* Second bissector diagonal gradient inside a rectangle *)
let grad_rect_d2 c1 c2 x y w h =
  let sc = w + h in
  let limx = x in
  let limy = y + h - _1_ in
  let rec loop i x0 y0 x1 y1 =
    set_color (scale_color i sc c1 c2);
    moveto x0 y0;
    if i <= sc then lineto x1 y1;
    if x0 > limx then
      if y1 < limy then
        loop (succ i) (pred x0) y0 x1 (succ y1) else
        loop (succ i) (pred x0) y0 (pred x1) y1 else
    if y0 < limy then
      if y1 < limy then
        loop (succ i) x0 (succ y0) x1 (succ y1) else
        loop (succ i) x0 (succ y0) (pred x1) y1 else
    () in
  set_line_width _1_;
  loop _0_ (x + w - _1_) y (x + w - _1_) y
;;

(************ Obsolete implementation using curve drawing primitives.
  It is OK for ratio of 1 because all isocolor curves are disjoint.
  When the ratio is not 1, isocolor curves may intersect, and may even be
  open.
  The easy and correct alternative is to use filling primitives that will do
  the trick automagically for us.
  Beware: now the drawing order is significative and we must draw from the
  border (associated with the [to_color]) to the center (associated with the
  [from_color]).

(* Generic radial gradient inside a rectangle
   (means growing balls with center xc yc). *)
let grad_rect_rad draw_ball c1 c2 xc yc rat x y w h =
  let xmin, xmax = x, x + w
  and ymin, ymax = y, y + h in
  let scx = min (xc - xmin) (xmax - xc)
  and scy = min (yc - ymin) (ymax - yc) in
  let sc = min scx scy in
  (** sc is the number of steps to perform. *)
  let rec loop r =
    if r < sc then begin
      set_color (scale_color r sc c1 c2);
      draw_ball xc yc r;
      loop (succ r);
    end in
  set_color c2;
  fill_rect x y w h;
  set_line_width _1_;
  loop _0_
;;

let draw_square x y r =
  draw_rect (x - r) (y - r) (2 * r + _1_) (2 * r + _1_)
;;

let grad_rect_rr = grad_rect_rad draw_square;;

let draw_circle x y r =
  if r = _0_ then plot x y else
  draw_circle x y r
;;

let grad_rect_re = grad_rect_rad draw_circle;;

(* Square radial gradient inside a rectangle
   (means growing squares with center xc yc). *)
let grad_rect_rr2 c1 c2 xc yc rat x y w h =
  let xmin, xmax = x, x + w
  and ymin, ymax = y, y + h in
  let scx = min (xc - xmin) (xmax - xc)
  and scy = min (yc - ymin) (ymax - yc) in
  let sc = min scx scy in
  (** sc is the number of steps to perform. *)
  let rec loop x0 y0 x1 y1 =
    if x1 >= xmax || y1 >= ymax ||
       x0 < xmin || y0 < ymin then () else begin
        moveto x0 y0;
        set_color (scale_color (yc - y0) sc c1 c2);
        lineto x1 y0;
        set_color (scale_color (x1 - xc) sc c1 c2);
        lineto x1 y1;
        set_color (scale_color (y1 - yc) sc c1 c2);
        lineto x0 y1;
        set_color (scale_color (xc - x0) sc c1 c2);
        lineto x0 y0;
        loop (pred x0) (pred y0) (x1 + 1) (succ y1);
      end in
  set_color c2;
  fill_rect x y w h;
  set_line_width _1_;
  loop xc yc xc yc
;;

(************ Obsolete implementation using curve drawing primitives. *)
*)

(* Generic radial gradient inside a rectangle
   (means growing balls with center xc yc). *)
let grad_rect_rad fill_ball c1 c2 xc yc (p, q) x y w h =
  let xmin, xmax = x, x + w
  and ymin, ymax = y, y + h in
  let scx = min (xc - xmin) (xmax - xc)
  and scy = min (yc - ymin) (ymax - yc) in
  let qsc = min (q * scx) (p * scy) in
  let sc  = qsc / q in
  (** sc is the number of steps to perform. *)
  let rec loop r =
    if r >= _0_ then begin
      set_color (scale_color r sc c1 c2);
      let qr = q * r in
      let rx = qr / q in
      let ry = qr / p in
      fill_ball xc yc rx ry;
      if r > _0_ then loop (pred r);
    end in
  set_color c2;
  fill_rect x y w h;
  set_line_width _1_;
  loop (pred sc)
;;

let grad_rect_rr = grad_rect_rad fill_ball_inf;;

let grad_rect_re = grad_rect_rad fill_ball_euclidean;;

(* Filling rectangles with gradients of colors. *)
let grad_rectangle gm x y w h =
  match gm with
  | Rect_Horizontal (c1, c2) ->
     grad_rect_h c1 c2 x y w h
  | Rect_Vertical (c1, c2) ->
     grad_rect_v c1 c2 x y w h
  | Rect_Diagonal1 (c1, c2) ->
     grad_rect_d1 c1 c2 x y w h
  | Rect_Diagonal2 (c1, c2) ->
     grad_rect_d2 c1 c2 x y w h
  | Rect_Radial_Rectangular (c1, c2, xc, yc, rat) ->
     grad_rect_rr c1 c2 xc yc rat x y w h
  | Rect_Radial_Elliptic (c1, c2, xc, yc, rat) ->
     grad_rect_re c1 c2 xc yc rat x y w h
;;

(* Filling squares with gradients of colors. *)
let grad_square gm x y l = grad_rectangle gm x y l l;;

(* Horizontal gradient inside an ellipse *)
let grad_ell_h c1 c2 x y w h =
  let rec loop i =
    if i < w then begin
      set_color (scale_color i w c1 c2);
      fill_rect (x + i) y _1_ h;
      loop (succ i);
    end in
  loop _0_
;;

(* Vertical gradient inside an ellipse *)
let grad_ell_v c1 c2 x y w h =
(* Was
  for i = _0_ to h - _1_ do
    set_color (scale_color i h c1 c2);
    fill_rect x (y + i) w _1_;
  done *)
 let rec loop i =
    if i < h then begin
      set_color (scale_color i h c1 c2);
      fill_rect x (y + i) w _1_;
      loop (succ i);
    end in
  loop _0_
;;

(* First bissector diagonal gradient inside an ellipse *)
let grad_ell_d1 c1 c2 x y w h =
  let sc = w + h in
  let limx = x + w - _1_ in
  let limy = y + h - _1_ in
  let rec loop i x0 y0 x1 y1 =
    set_color (scale_color i sc c1 c2);
    moveto x0 y0;
    if i <= sc then lineto x1 y1;
    if x0 < limx then
      if y1 < limy then
        loop (succ i) (succ x0) y0 x1 (succ y1) else
        loop (succ i) (succ x0) y0 (succ x1) y1 else
    if y0 < limy then
      if y1 < limy then
        loop (succ i) x0 (succ y0) x1 (succ y1) else
        loop (succ i) x0 (succ y0) (succ x1) y1 else
    () in
  set_line_width _1_;
  loop _0_ x y x y
;;

(* Second bissector diagonal gradient inside an ellipse *)
let grad_ell_d2 c1 c2 x y w h =
  let sc = w + h in
  let limx = x in
  let limy = y + h - _1_ in
  let rec loop i x0 y0 x1 y1 =
    set_color (scale_color i sc c1 c2);
    moveto x0 y0;
    if i <= sc then lineto x1 y1;
    if x0 > limx then
      if y1 < limy then
        loop (succ i) (pred x0) y0 x1 (succ y1) else
        loop (succ i) (pred x0) y0 (pred x1) y1 else
    if y0 < limy then
      if y1 < limy then
        loop (succ i) x0 (succ y0) x1 (succ y1) else
        loop (succ i) x0 (succ y0) (pred x1) y1 else
    () in
  set_line_width _1_;
  loop _0_ (x + w - _1_) y (x + w - _1_) y
;;

(*
(* Generic radial gradient inside an ellipse
   (means growing balls with center xc yc). *)
let grad_ell_rad fill_ball c1 c2 xc yc (p, q) x y rx ry =
  let xmin, xmax = x, x + w
  and ymin, ymax = y, y + h in
  let scx = min (xc - xmin) (xmax - xc)
  and scy = min (yc - ymin) (ymax - yc) in
  let qsc = min (q * scx) (p * scy) in
  let sc  = qsc / q in
  (** sc is the number of steps to perform. *)
  let rec loop r =
    if r >= _0_ then begin
      set_color (scale_color r sc c1 c2);
      let qr = q * r in
      let rx = qr / q in
      let ry = qr / p in
      fill_ball xc yc rx ry;
      loop (pred r);
    end in
  set_color c2;
  fill_rect x y w h;
  set_line_width _1_;
  loop (pred sc)
;;
*)

(* Rectangular radial gradient inside an ellipse
   (means growing rectangles with center xc yc). *)
let grad_ell_rr c1 c2 xc yc _rat x y w h =
  let xmin, xmax = x, x + w
  and ymin, ymax = y, y + h in
  let scx = min (xc - xmin) (xmax - xc)
  and scy = min (yc - ymin) (ymax - yc) in
  let sc = min scx scy in
  (** sc is the number of steps to perform. *)
  let rec loop x0 y0 x1 y1 =
    if x1 >= xmax || y1 >= ymax ||
       x0 < xmin || y0 < ymin then () else begin
        moveto x0 y0;
        set_color (scale_color (yc - y0) sc c1 c2);
        lineto x1 y0;
        set_color (scale_color (x1 - xc) sc c1 c2);
        lineto x1 y1;
        set_color (scale_color (y1 - yc) sc c1 c2);
        lineto x0 y1;
        set_color (scale_color (xc - x0) sc c1 c2);
        lineto x0 y0;
        loop (pred x0) (pred y0) (succ x1) (succ y1);
      end in
  set_color c2;
  fill_rect x y w h;
  set_line_width _1_;
  loop xc yc xc yc
;;

(* Elliptic radial gradient inside an ellipse
   (means growing ellipses with center xc yc). *)
let grad_ell_re c1 c2 xc yc _rat x y w h =
  let xmin, xmax = x, x + w
  and ymin, ymax = y, y + h in
  let scx = min (xc - xmin) (xmax - xc)
  and scy = min (yc - ymin) (ymax - yc) in
  let sc = min scx scy in
  (** sc is the number of steps to perform. *)
  let rec loop r =
    if r < sc then begin
      set_color (scale_color r sc c1 c2);
      draw_circle xc yc r;
      loop (succ r);
    end in
  set_color c2;
  fill_rect x y w h;
  if sc > _0_ then (set_color c1; plot xc yc);
  set_line_width _1_;
  loop _0_
;;

(* Filling ellipses with gradients of colors. *)
let grad_ellipse gm x y rx ry =
  match gm with
  | Ellipse_Horizontal (c1, c2) ->
     grad_ell_h c1 c2 x y rx ry
  | Ellipse_Vertical (c1, c2) ->
     grad_ell_v c1 c2 x y rx ry
  | Ellipse_Diagonal1 (c1, c2) ->
     grad_ell_d1 c1 c2 x y rx ry
  | Ellipse_Diagonal2 (c1, c2) ->
     grad_ell_d2 c1 c2 x y rx ry
  | Ellipse_Radial_Rectangular (c1, c2, xc, yc, rat) ->
     grad_ell_rr c1 c2 xc yc rat x y rx ry
  | Ellipse_Radial_Elliptic (c1, c2, xc, yc, rat) ->
     grad_ell_re c1 c2 xc yc rat x y rx ry
;;

(* Filling circles with gradients of colors. *)
let grad_circle gm x y r = grad_ellipse gm x y r r;;

(* Filling arcs with gradients of colors. *)
let grad_arc gm x y rx ry a1 a2 =
  let gmax, c, c' =
    match gm with
    | Arc_Horizontal (c, c') -> rx, c, c'
    | Arc_Vertical (c, c') -> ry, c, c'
    | Arc_Diagonal1 (c, c') -> (rx + ry + _1_) / _2_, c, c'
    | Arc_Diagonal2 (c, c') -> (rx + ry + _1_) / _2_, c, c'
    | Arc_Circular (c, c') -> max rx ry, c, c' in
  (* Since
     - draw_arc x y 0 ry does something,
     - draw_arc x y rx 0 does something,
     - draw_arc x y 0 0 does nothing,
     - draw_arc x y 1 1 draws a square with _1_ point cleared,
     we must prevent rx, ry to be successively (0, 0) then (1, 1), otherwise
     _1_ single point would be left cleared at the very center of the arc.
     Hence rxmin is 1 if rx is not 0, and rymin is 1 if rxmin = 0! *)
  let rxmin = if rx > _0_ then _1_ else _0_ in
  let rymin = if rxmin > _0_ then _0_ else _1_ in
  let scale_rx =
    match gm with
    | Arc_Horizontal (_, _) -> (fun i -> i)
    | Arc_Vertical (_, _) -> (fun _i -> rx)
    | Arc_Diagonal1 (_, _) -> (fun _i -> rx)
    | Arc_Diagonal2 (_, _) -> (fun i -> i)
    | Arc_Circular (_, _) -> (fun i -> scale i gmax rxmin rx) in
  let scale_ry =
    match gm with
    | Arc_Horizontal (_, _) -> (fun _i -> ry)
    | Arc_Vertical (_, _) -> (fun i -> i)
    | Arc_Diagonal1 (_, _) -> (fun i -> i)
    | Arc_Diagonal2 (_, _) -> (fun _i -> ry)
    | Arc_Circular (_, _) -> (fun i -> scale i gmax rymin ry) in
(*
    for i = _0_ to gmax do
      set_color (scale_color i gmax c c');
      let rx = scale_rx i
      and ry = scale_ry i in
      draw_arc x y rx ry a1 a2;
    done
*)
    let rec loop i =
      if i < gmax then begin
        set_color (scale_color i gmax c c');
        let rx = scale_rx i
        and ry = scale_ry i in
        draw_arc x y rx ry a1 a2;
        loop (succ i);
      end in
    loop _0_
;;
